#============================================
# To select video frames with trackbar
#============================================


#============================================
# User input
#============================================

# Scale conversion to convert pixel to cm
SCALE_CONVERSION = 0.264   # cm per pixel, compute as length_cm = length_pixel * conversion factor

# Enter window name
WINDOW_NAME = "Exhaustion analyser"

# Enter default video playing speed
VIDEO_SPEED = 0.02

TRACKBAR_NAME = 'Frame'

CURRENT_FRAME = 1

#============================================
# Import libraries
#============================================

import cv2
import os
import glob
import sys
import csv
import time
import tkinter as tk
from tkinter import filedialog
import pandas as pd


#=========================================================================================================
# Working directory and file paths
#=========================================================================================================
# path = "

# Create Tkinter window dialog
root = tk.Tk()
root.withdraw()

# Ask user to select working directory
DIRECTORY_PATH = filedialog.askdirectory()
# DIRECTORY_PATH = "D:/Research/Projects/Project_RockLobster/Temperature acclimation/Exhaustion/Analysis/ExhaustionAnalyser"

# Now change the directory
os.chdir(DIRECTORY_PATH)
print(DIRECTORY_PATH , "set as working directory")

# Ask user for video file
FILE_PATH = filedialog.askopenfilename()
print("Video file", FILE_PATH, " opened")


#============================================
# Video player functions
#============================================

############################################################
# Set video capture frame to trackbar value
def onChange(trackbarValue):
    # Set the selected frame
    cap.set(cv2.CAP_PROP_POS_FRAMES,trackbarValue)
    err,frame = cap.read()
    # Update the window
    cv2.imshow(WINDOW_NAME, frame)

    pass


# Display border and frame below video
def frame_display(frame):
    # Create display frame below video
    frame = cv2.copyMakeBorder(frame, 0, 40, 0, 0, cv2.BORDER_CONSTANT, value=[0,0,0]);

    # Display current frame below video
    cv2.putText(frame, str("Current frame: " + str(CURRENT_FRAME) + "/" + str(video_length)), (5, 375), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255, 255, 255), 1, cv2.LINE_AA)

    return frame


############################################################
# Function generator to access functions via key press
def skip_frame_generator(skipped_frames):

    def skip_frame():

        # get the current frame position
        current_frame = cap.get(cv2.CAP_PROP_POS_FRAMES) - 1

        # Only skip frames if this will not exceed total video length
        if current_frame + skipped_frames < video_length:
            # skip of df frames
            cap.set(cv2.CAP_PROP_POS_FRAMES, current_frame + skipped_frames)
            # update the trackbar position
            cv2.setTrackbarPos(TRACKBAR_NAME, WINDOW_NAME, int(cap.get(cv2.CAP_PROP_POS_FRAMES)))
        # read and update the frame
        _, frame = cap.read()

    return skip_frame


############################################################
# Increase video play speed
def video_faster():
    global VIDEO_SPEED
    VIDEO_SPEED = round(VIDEO_SPEED/2, 2)
    print(VIDEO_SPEED)

############################################################
# Decrease video play speed
def video_slower():
    global VIDEO_SPEED
    VIDEO_SPEED = round(VIDEO_SPEED * 2, 2)
    print(VIDEO_SPEED)

############################################################
# Play pause function
def play_pause():

    while True:

        # Intitiate key press variable
        key2 = cv2.waitKey(1) or 0xff

        # Press p to pause video
        if key2 == ord(' '):
            break


############################################################
# Quit player
def quit():
    # When everything done, release the capture
    cap.release()
    cv2.destroyAllWindows()
    print("Video analysis closed")
    exit(0)


############################################################
# Define default action (nothing) for action dictionary
def dummy():
    pass


############################################################
# Display recorded tail flip

def display_closest_record(frame_records, current_frame, label, pos_x, pos_y):

    if len(frame_records) > 0:
        # Find data record closest to current frame
        closest_data_record = min(frame_records, key=lambda x: abs(x - current_frame))

        # Display closest tail flip in green if equal to current frame
        if current_frame in frame_records:
            cv2.putText(frame, str(label + ": " + str(closest_data_record)), (pos_x, pos_y), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (0, 255, 0), 1,
                        cv2.LINE_AA)
        # Otherwise in white
        else:
            cv2.putText(frame, str(label + ": " + str(closest_data_record)), (pos_x, pos_y), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255, 255, 255), 1,
                        cv2.LINE_AA)


#============================================
# File loading/saving functions
#============================================

# Class to load or save files
class File:
    def __init__(self):
        self.file_name = ""
        self.data = []

    def get_filename(self, file_path):
        self.file_name = os.path.basename(file_path)

        # Extract file name without file extension
        self.file_name = os.path.splitext(self.file_name)[0]

        print(self.file_name)

    def load_csv(self, file_label):
        # Create csv file name
        csv_file_name = self.file_name + "_" + file_label + '.csv'

        # Check if data file exists
        if glob.glob(csv_file_name):
            # If yes, load data
            self.data = pd.read_csv(csv_file_name)

            # Show loaded data
            print(self.data)

            # Confirm data were loaded
            print("Existing data loaded from " + csv_file_name)
        else:  # If data do not exist
            print("No file named", csv_file_name)

        # Return data variable
        return self.data

    # Save data as pandas data frame
    def save_csv(self, data_class, file_label):
        # Internal function to allow passing parameters e.g. file_label in key press function

        def save_csv_internal():
            # Check if escape or other records are saved as this affects the file format
            if len(list(data_class.get_record())) > 2:
                # Get records
                data_frame, data_pos_x, data_pos_y, data_action = data_class.get_record()
                # Save list data in dictionary
                data_dict = {"Frame": data_frame,
                             "pos_x": data_pos_x,
                             "pos_y": data_pos_y,
                             file_label: data_action,
                             "Conversion_scale": [SCALE_CONVERSION] * len(data_frame)}
            else:
                # Get records
                data_frame, data_action = data_class.get_record()

                # Save list data in dictionary
                data_dict = {"Frame": data_frame, file_label: data_action}

            # Convert dictionary to pandas data frame
            pandas_df = pd.DataFrame(data_dict)

            # Create csv file name
            csv_file_name = self.file_name + "_" + file_label + '.csv'

            # Save the data frame using the video file name
            pandas_df.to_csv(csv_file_name, index=False)

            # Inform user that data were saved
            print("Data saved in ", csv_file_name)

        return save_csv_internal


#============================================
# Analysis functions
#============================================

############################################################
# Record actions

class Record:
    def __init__(self, data_dict):
        # Extract dictionary columns into list
        if len(data_dict) > 0:
            self.frame = data_dict.iloc[:, 0].tolist()  # iloc allows access of column by index
            self.action = data_dict.iloc[:, 1].tolist()  # iloc allows access of column by index
        else:
            self.frame = []
            self.action = []

    # Record data
    def record(self, action_label):
        # Internal function to allow passing parameters e.g. file_label in key press function
        def record_internal():
            # Appends frame number at which righting starts
            self.frame.append(CURRENT_FRAME)
            self.action.append(action_label)

            # Display last records
            print(len(self.frame), self.frame[-1:], self.action[-1:])

        return record_internal

    # Function to delete last righting record
    def delete(self):

        if len(self.frame) >= 1:
            # Display last records
            print(len(self.frame), self.frame[-1:], self.action[-1:], " deleted")
            # Delete last righting record
            self.frame.pop(-1)
            self.action.pop(-1)
        else:
            print("No records")

    # Return frame and action records
    def get_record(self):
        return self.frame, self.action

    def display_record(self, label, pos_x, pos_y):
        display_closest_record(self.frame, CURRENT_FRAME, label, pos_x, pos_y)


############################################################
# Record escape speed

class escape():

    def __init__(self, data_dict):
        # Extract dictionary columns into list if already existing
        if len(data_dict) > 0:
            self.frame = data_dict.iloc[:, 0].tolist()  # Read frame number
            self.pos_x = data_dict.iloc[:, 1].tolist()  # Read x coordinate
            self.pos_y = data_dict.iloc[:, 2].tolist()  # Read y coordinate
            self.action = data_dict.iloc[:, 3].tolist()  # Read action label
            self.conversion = data_dict.iloc[:, 4].tolist()  # Read conversion scale
        else:
            self.frame = []  # Variable to store frame number
            self.pos_x = []  # Variable to store xy coordinates
            self.pos_y = []  # Variable to store xy coordinates
            self.action = []  # Record if this was beginning or end of escape jet
            self.conversion = []  # Conversion scale pixel to cm

    def record(self, event,x,y, flags, param):
        # Record x y coordinates when left mouse button is pressed
        if event == cv2.EVENT_LBUTTONDOWN:
            # Draw circle on action
            cv2.circle(frame,(x,y),50,(0,0,255),-1)

            # Append frame number and xy coordinates
            self.frame.append(CURRENT_FRAME)
            self.pos_x.append(x)
            self.pos_y.append(y)
            self.action.append("Start")

            # Print record
            print(len(self.frame), self.frame[-1:], self.pos_x[-1:], self.pos_y[-1:], self.action[-1:])

        # Record x y coordinates when left mouse button is released
        elif event is cv2.EVENT_RBUTTONDOWN:
            # Draw circle on action
            cv2.circle(frame, (x, y), 50, (0, 255, 0), -1)

            # Append frame number and xy coordinates
            self.frame.append(CURRENT_FRAME)
            self.pos_x.append(x)
            self.pos_y.append(y)
            self.action.append("End")

            # Print record
            print(len(self.frame), self.frame[-1:], self.pos_x[-1:], self.pos_y[-1:], self.action[-1:])

    def delete(self):

        if len(self.frame) >= 1:
            # Display last records
            print(len(self.frame), self.frame[-1:], self.pos_x[-1:], self.pos_y[-1:], self.action[-1:], " deleted")

            # Delete last righting record
            self.frame.pop(-1)
            self.pos_x.pop(-1)
            self.pos_y.pop(-1)
            self.action.pop(-1)

        else:
            print("No records")

    # Return frame and action records
    def get_record(self):
        return self.frame, self.pos_x, self.pos_y, self.action

    def display_record(self, label, pos_x, pos_y):
        display_closest_record(self.frame, CURRENT_FRAME, label, pos_x, pos_y)


#============================================
# Single call commands
#============================================

############################################################
# Instantiate classes and load data

# Instantiate File class
file = File()
# Extract video filename
file.get_filename(FILE_PATH)

# Load data if they already exist
# Tail flips
tail_flip_data = file.load_csv("TailFlips")
# Rightings
righting_data = file.load_csv("Rightings")
# Escape speed
escape_data = file.load_csv("EscapeSpeed")

# Instantiate class to record data
tail_flip = Record(tail_flip_data)
righting = Record(righting_data)
escape = escape(escape_data)


############################################################
# Build dictionary with functions called by key press
actions = dict()

# Video player
actions[ord(" ")] = play_pause
actions[ord("q")] = quit
#Frame control
actions[ord("4")] = skip_frame_generator(300)
actions[ord("3")] = skip_frame_generator(40)
actions[ord("2")] = skip_frame_generator(-50)
actions[ord("1")] = skip_frame_generator(-400)
# Video speed
actions[ord("f")] = video_faster # Function either be stored with lambda or without brackets
actions[ord("s")] = video_slower

# Tail flips
actions[ord("t")] = tail_flip.record("TailFlip")
actions[ord("r")] = tail_flip.delete
actions[ord("8")] = file.save_csv(tail_flip, "TailFlips")
# Righting
actions[ord("p")] = righting.record("Righting_start")
actions[ord("o")] = righting.record("Righting_end")
actions[ord("i")] = righting.delete
actions[ord("9")] = file.save_csv(righting, "Rightings") # needs class and label
# Escape speed
actions[ord("0")] = file.save_csv(escape, "EscapeSpeed") # needs class and label
actions[ord("m")] = escape.delete # needs class and label

# Capture video from file
cap = cv2.VideoCapture(FILE_PATH)

# Get total number of video frames
video_length = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))

# Get video frame rate
fps = round(cap.get(cv2.CAP_PROP_FPS), 0)
print("Video frame rate is ", fps, "frames per second")

# Create window
cv2.namedWindow(WINDOW_NAME)

# Create trackbar object
cv2.createTrackbar(TRACKBAR_NAME, WINDOW_NAME, 0, video_length, onChange )

# Set first video frame to 0
onChange(0)

# Start after pressing any key
cv2.waitKey()


#============================================
# Main loop
#============================================

while cap.isOpened():

    # Read video image frame by frame
    ret,frame = cap.read()

    # Save current frame
    CURRENT_FRAME = int(cap.get(cv2.CAP_PROP_POS_FRAMES))

    # Save pressed key
    key = cv2.waitKey(10) & 0xff

    # Perform action depending on key press
    actions.get(key, dummy)()

    cv2.setMouseCallback(WINDOW_NAME, escape.record)

    # Create display frame below video
    frame = frame_display(frame)

    # Display records
    tail_flip.display_record("Tail Flip", 200, 375)
    righting.display_record("Righting", 200, 395)
    escape.display_record("Escape", 400, 375)

    # Display video image in window
    cv2.imshow(WINDOW_NAME, frame)

    # Save data if last frame is reached
    if CURRENT_FRAME == video_length-1:
        file.save_csv(escape, "EscapeSpeed")()
        quit()

    # Waiting time to control video play speed
    time.sleep(VIDEO_SPEED)

